# SPDX-FileCopyrightText: 2021 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: LGPL-2.0-or-later

from config import *
import datetime
import functools
import os
import pytz
import time
from qgis import *
from qgis.core import *

xStep = xRange / (1 << zDepth)
yStep = yRange / (1 << zDepth)

def z2x(z):
    x = 0
    for i in range(0, zDepth):
        x += (z & (1 << i * 2)) >> i
    return x

def z2y(z):
    y = 0
    for i in range(0, zDepth):
        y += (z & (1 << (1 + i * 2))) >> (i + 1)
    return y

def rectForZ(z, depth):
    mask = (1 << (2*(zDepth - depth))) - 1
    x = z2x(z & ~mask) * xStep + xStart
    y = z2y(z & ~mask) * yStep + yStart
    xSize = xRange / (1 << depth)
    ySize = yRange / (1 << depth)
    return QgsRectangle(x, y, x + xSize, y + ySize)

def tzIdToEnum(tzId):
    if not tzId:
        return 'Undefined'
    return tzId.replace('/', '_').replace('-', '_')

# Layer specific configuration for things considered in the spatial index
class LayerConfig:
    def readFeatureValue(self, feature):
        return feature[self.featureKey]

    def isValidFeature(self, feature):
        return feature != None

    def normalizeAndFilter(self, features):
        for key in features:
            if not self.isValidFeature(key):
                del(features[key])
        return features

    def resolveConflicts(self, features):
        out = []
        for key in features:
            out.append((key, features[key]));
        n = functools.reduce(lambda n, f: n + f[1], out, 0)
        if n > 0:
            out = [(k, v/n) for (k, v) in out]
        out = list(filter(lambda x: x[1] > featureAreaRatioThreshold, out))
        out.sort(key = lambda x: x[1], reverse = True)
        return out

class TzLayerConfig(LayerConfig):
    def __init__(self):
        self.layer = 'tzLayer'
        self.featureKey = 'tzid'

    def readFeatureValue(self, feature):
        tzId = feature[self.featureKey]
        if tzId in TZID_MAP:
            return TZID_MAP[tzId]
        return tzId

    def resolveConflicts(self, features):
        if len(features) > 1:
            tzs = list(features.keys())
            # check for conflicting timezones
            tz = pytz.timezone(tzs[0])
            if not all(self.isSameTimezone(tz, pytz.timezone(x)) for x in tzs[1:]):
                return None
            out = super().resolveConflicts(features)
            return [out[0]]
        return super().resolveConflicts(features)

    def isSameTimezone(self, lhs, rhs):
        try:
            # hacky tz comparison, lacking access to the rules for comparing actual DST transition times
            # TODO stabilize results and ensure we capture differences due to different week boundaries
            dt = datetime.datetime.today().toordinal()
            return all(lhs.utcoffset(datetime.datetime.fromordinal(dt + 30*x)) == rhs.utcoffset(datetime.datetime.fromordinal(dt + 30*x))
                   and lhs.tzname(datetime.datetime.fromordinal(dt + 30*x)) == rhs.tzname(datetime.datetime.fromordinal(dt + 30*x)) for x in range(0, 11))
        except:
            return False

class CountryLayerConfig(LayerConfig):
    def __init__(self):
        self.layer = 'countryLayer'
        self.featureKey = 'ISO3166-1'

    def readFeatureValue(self, feature):
        code = feature[self.featureKey]
        if code in ISO3166_1_MAP:
            return ISO3166_1_MAP[code]
        return code

    def normalizeAndFilter(self, features):
        if len(features) > 1:
            # apply country disambiguation
            for d in ISO3166_1_DISAMBIGUATION_MAP:
                if d[0] in features and d[1] in features:
                    del(features[d[1]])
        return super().normalizeAndFilter(features)

    def resolveConflicts(self, features):
        out = super().resolveConflicts(features)
        if len(out) > 1:
            return None
        return out

class SubdivLayerConfig(LayerConfig):
    def __init__(self):
        self.layer = 'subdivLayer'
        self.featureKey = 'ISO3166-2'

    def resolveConflicts(self, features):
        out = super().resolveConflicts(features)
        if len(out) > 1:
            return None
        return out
#
# Parallelized spatial index computation of a single sub-tile
#
class SpatialIndexerSubTask(QgsTask):
    def __init__(self, context, zStart, zStartDepth):
        super().__init__('Compute spatial index sub-tile ' + str(zStart), QgsTask.CanCancel)
        self.context = context
        self.zStart = zStart
        self.zStartDepth = zStartDepth
        self.lastFeature = []
        self.exception = None
        self.result = []
        self.layerConfig = [TzLayerConfig(), CountryLayerConfig(), SubdivLayerConfig()]

    def run(self):
        try:
            self.computeTile(self.zStart, self.zStartDepth)
        except Exception as e:
            self.exception = e
            QgsMessageLog.logMessage('Exception in task "{}"'.format(self.exception), LOG_CATEGORY, Qgis.Info)
            return False
        return True

    def computeTile(self, zStart, depth):
        if self.isCanceled() or depth < 1:
            return
        z = zStart
        d = depth - 1
        zIncrement = 1 << (2*d)
        for i in range(0, 4):
            # find features in the input vector layer inside our current tile
            # we cannot take shortcuts here and avoid the expensive area computation
            # as we do get spurious 0-area results for some large and disjoint polygons (FR, US, NZ, etc)
            feature = []
            featureCount = 0
            for layerConfig in self.layerConfig:
                l = {}
                rect = rectForZ(z, zDepth - d)
                rectGeo = QgsGeometry.fromRect(rect)
                for f in self.context[layerConfig.layer].getFeatures(rect):
                    featureArea = f.geometry().intersection(rectGeo).area()
                    if featureArea > 0.0:
                        key = layerConfig.readFeatureValue(f)
                        if key in l:
                            l[key] += featureArea / rectGeo.area()
                        else:
                            l[key] = featureArea / rectGeo.area()
                feature.append(layerConfig.normalizeAndFilter(l))
                featureCount = max(featureCount, len(l))

            # recurse on conflicts
            if depth > 1 and featureCount > 1:
                self.computeTile(z, d)
            # leaf tile, process the result
            else:
                # try to clean up any remaining conflicts
                for i in range(len(feature)):
                    feature[i] = self.layerConfig[i].resolveConflicts(feature[i])
                self.resolveConflicts(feature)
                # if there's a change to the previous value, propagate to the result output
                if self.lastFeature != feature and feature != []:
                    self.result.append((z, feature))
                    self.lastFeature = feature

            z += zIncrement

    def resolveConflicts(self, feature):
        # if we have a unique subdivision but no country, use the subdivision's country
        # this happens in coastal regions when territorial waters overlap in the tile, but
        # actual land boundaries don't. As we primarily care about correct result on land,
        # we can ignore the overlap on water
        if (not feature[1] or len(feature[1]) == 0) and feature[2] and len(feature[2]) == 1:
            feature[1] = [(feature[2][0][0][:2], 1)]

        # if subdivision country and country code conflict, discard the subdivision
        # this is mainly a problem for French overseas territories, and there the country
        # code is actually more useful
        if feature[1] and len(feature[1]) == 1 and feature[2] and len(feature[2]) == 1 and feature[1][0][0] != feature[2][0][0][:2]:
            feature[2] = None

    def finished(self, result):
        if not result and self.exception != None:
            QgsMessageLog.logMessage('Task "{name}" Exception: {exception}'.format(name=self.description(), exception=self.exception), LOG_CATEGORY, Qgis.Critical)
            raise self.exception

#
# Task for spawning the sub-tasks doing the actual work, and accumulating the result
#
class SpatialIndexerTask(QgsTask):
    def __init__(self, context, loadLayersTask):
        super().__init__('Compute spatial index', QgsTask.CanCancel)
        self.context = context
        self.tasks = []
        self.conflictTiles = [0, 0, 0]
        self.startTime = time.time()
        self.propertyTable = []

        startDepth = 4
        startIncrement = 1 << (2 * (zDepth - startDepth))
        for i in range(0, (1 << (2 * startDepth))):
            task = SpatialIndexerSubTask(context, i * startIncrement, zDepth - startDepth)
            self.addSubTask(task, [loadLayersTask], QgsTask.ParentDependsOnSubTask)
            self.tasks.append(task)

    def run(self):
        try:
            QgsMessageLog.logMessage('Aggregating results...', LOG_CATEGORY, Qgis.Info)

            # compact the spatial index and prepare the property map
            spatialIndex = []
            propertyMap = {}
            prevProperty = (None, None, None)
            propertyMap[prevProperty] = 0
            for task in self.tasks:
                for (z,res) in task.result:
                    res = self.normalize(res);
                    tmp = []
                    for i in range(len(res)):
                        if res[i] == None or len(res[i]) == 0:
                            tmp.append(None)
                        else:
                            tmp.append(res[i][0][0])
                        if res[i] == None:
                            self.conflictTiles[i] += 1
                    prop = (tmp[0], tmp[1], tmp[2])

                    if prevProperty == prop:
                        continue
                    prevProperty = prop
                    spatialIndex.append((z, prop))
                    propertyMap[prop] = 0

            for prop in propertyMap:
                self.propertyTable.append(prop)
            self.propertyTable.sort(key = lambda x: (x[0] if x[0] else "", x[1] if x[1] else "", x[2] if x[2] else ""))
            for i in range(len(self.propertyTable)):
                propertyMap[self.propertyTable[i]] = i

            # write spatial index
            out = open('../../data/spatial_index_data.cpp', 'w')
            out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated spatial index generated using QGIS.
 */

#include "spatial_index_entry_p.h"

static constexpr const SpatialIndexEntry spatial_index[] = {
    // clang-format off
""")
            prevProperties = (None, None, None)
            for (z, prop) in spatialIndex:
                out.write(f"    {{{z}, {propertyMap[prop]}}},\n")
            out.write("    // clang-format on\n};\n")
            out.close()

            # write property table
            out = open('../../data/spatial_index_properties.cpp', 'w')
            out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated spatial index generated using QGIS.
 */

#include "spatial_index_property_p.h"
#include "timezone_names_p.h"

static constexpr const SpatialIndexProperty spatial_index_properties[] = {
""")
            for p in self.propertyTable:
                if not p[1] and not p[2]:
                    out.write(f"    {{Tz::{tzIdToEnum(p[0])}}},\n")
                elif not p[2]:
                    out.write(f"    {{Tz::{tzIdToEnum(p[0])}, \"{p[1]}\"}},\n")
                else:
                    out.write(f"    {{Tz::{tzIdToEnum(p[0])}, \"{p[2]}\"}},\n")
            out.write("};\n")
            out.close()

            # write zindex parameters
            out = open('../../data/spatial_index_parameters_p.h', 'w')
            out.write("""/*
 * SPDX-License-Identifier: CC0-1.0
 * SPDX-FileCopyrightText: none
 *
 * Autogenerated spatial index generated using QGIS.
 */

#include <cstdint>

""")
            out.write(f"constexpr const float XStart = {xStart};\n")
            out.write(f"constexpr const float XRange = {xRange};\n")
            out.write(f"constexpr const float YStart = {yStart};\n")
            out.write(f"constexpr const float YRange = {yRange};\n")
            out.write(f"constexpr const uint8_t ZDepth = {zDepth};\n")
            out.close()

            return True

        except Exception as e:
            self.exception = e
            QgsMessageLog.logMessage('Exception in task "{}"'.format(self.exception), LOG_CATEGORY, Qgis.Info)
            return False

    def normalize(self, feature):
        # normalization we couldn't do in the sub tasks as they rely on results not available yet at that point
        # fill in missing timezones from country/subdiv to tz maps
        # this is needed for the same reason we do the subdivision to country back filling above, conflicts in
        # territorial waters but a unique result on land
        tz = None
        if feature[2] and len(feature[2]) == 1:
            subdivCode = feature[2][0][0]
            subdivToTz = self.context['subdivisionToTimezoneMap']
            if subdivCode in subdivToTz and len(subdivToTz[subdivCode]) == 1:
                tz = list(subdivToTz[subdivCode])[0]
        if not tz and feature[1] and len(feature[1]) == 1:
            countryCode = feature[1][0][0]
            countryToTz = self.context['countryToTimezoneMap']
            if countryCode in countryToTz and len(countryToTz[countryCode]) == 1:
                tz = list(countryToTz[countryCode])[0]
        if tz:
            feature[0] = [(tz, 1)]

        return feature

    def finished(self, result):
        QgsMessageLog.logMessage('Finished task "{}"'.format(self.description()), LOG_CATEGORY, Qgis.Info)
        tileCount = 1 << (2 * zDepth)
        for i in range(len(self.conflictTiles)):
            QgsMessageLog.logMessage(f" {self.conflictTiles[i] / tileCount} of feature {i} area is conflicted", LOG_CATEGORY, Qgis.Info)
        QgsMessageLog.logMessage(f" collected {len(self.propertyTable)} distinct features", LOG_CATEGORY, Qgis.Info)
        QgsMessageLog.logMessage(f" computation took {time.time() - self.startTime} seconds", LOG_CATEGORY, Qgis.Info)
        if not result and self.exception != None:
            QgsMessageLog.logMessage('Task "{name}" Exception: {exception}'.format(name=self.description(), exception=self.exception), LOG_CATEGORY, Qgis.Critical)
            raise self.exception




# SPDX-FileCopyrightText: 2021 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: LGPL-2.0-or-later

# data sources
TZDATA_VERSION = '2020d'
TZDATA_URL = f"https://github.com/evansiroky/timezone-boundary-builder/releases/download/{TZDATA_VERSION}/timezones.shapefile.zip"
ISO3166_1_VERSION = '2021-08-16'
ISO3166_1_URL = f"https://volkerkrause.eu/~vkrause/iso3166-boundaries/iso3166-1-boundaries.shp-{ISO3166_1_VERSION}.zip"
ISO3166_2_VERSION = ISO3166_1_VERSION
ISO3166_2_URL = f"https://volkerkrause.eu/~vkrause/iso3166-boundaries/iso3166-2-boundaries.shp-{ISO3166_2_VERSION}.zip"

# ISO 3166-1 code mappings
# This is for codes that should be unconditionally replaced, which is mainly useful
# for certain historical or political corner-cases or "non-countries" like Antarctica
ISO3166_1_MAP = {
    'AQ': None,
    'AX': 'FI',
    'SJ': 'NO',
    'UM': 'US'
}

# ISO 3166-1 overlap disambiguation
# This is mainly relevant for oversees territories that both have their own code
# as well that of the country they belong to
ISO3166_1_DISAMBIGUATION_MAP = {
    ( 'AS', 'US' ),
    ( 'AW', 'NL' ),
    ( 'BL', 'FR' ),
    ( 'BQ', 'NL' ),
    ( 'BT', 'CN' ),
    ( 'CC', 'AU' ),
    ( 'CW', 'NL' ),
    ( 'CX', 'AU' ),
    ( 'GF', 'FR' ),
    ( 'GP', 'FR' ),
    ( 'GU', 'US' ),
    ( 'HK', 'CN' ),
    ( 'HT', 'US' ),
    ( 'MF', 'FR' ),
    ( 'MO', 'CN' ),
    ( 'MP', 'US' ),
    ( 'MQ', 'FR' ),
    ( 'NC', 'FR' ),
    ( 'NF', 'AU' ),
    ( 'PF', 'FR' ),
    ( 'PM', 'FR' ),
    ( 'PR', 'US' ),
    ( 'RE', 'FR' ),
    ( 'SX', 'NL' ),
    ( 'TF', 'FR' ),
    ( 'VI', 'US' ),
    ( 'WF', 'FR' ),
    ( 'YT', 'FR' )
}

# ISO 3166-2 filter configuration
# most of those are sub-subdivisions
ISO3166_2_FILTER = [
    { 'country': 'BD', 'admin_level': 5 },
    { 'country': 'BE', 'admin_level': 6 },
    { 'country': 'BF', 'admin_level': 5 },
    { 'country': 'CZ', 'admin_level': 7 },
    { 'country': 'ES', 'admin_level': 6 },
    { 'country': 'FR', 'admin_level': 5 },
    { 'country': 'FR', 'admin_level': 6 },
    { 'country': 'GB', 'admin_level': 6 },
    { 'country': 'GB', 'admin_level': 8 },
    { 'country': 'GN', 'admin_level': 6 },
    { 'country': 'GQ', 'admin_level': 4 },
    { 'country': 'GW', 'admin_level': 4 },
    { 'country': 'IE', 'admin_level': 6 },
    { 'country': 'IE', 'admin_level': 7 },
    { 'country': 'IT', 'admin_level': 6 },
    { 'country': 'LK', 'admin_level': 5 },
    { 'country': 'LT', 'admin_level': 5 },
    { 'country': 'MA', 'admin_level': 5 },
    { 'country': 'MW', 'admin_level': 4 },
    { 'country': 'NP', 'admin_level': 4 },
    { 'country': 'NP', 'admin_level': 5 },
    { 'country': 'PH', 'admin_level': 4 },
    { 'country': 'UG', 'admin_level': 4 }
]


# Timezone mapping
# use this to replace timezone ids with an equivalent
# this is mainly useful for filtering out timezones of limited practical or merely historical relevance,
# trading historical accuracy for practical usability for current and future date/time values
TZID_MAP = {
    'America/Argentina/Catamarca': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Cordoba': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Jujuy': 'America/Argentina/Buenos_Aires',
    'America/Argentina/La_Rioja': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Mendoza': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Rio_Gallegos': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Salta': 'America/Argentina/Buenos_Aires',
    'America/Argentina/San_Juan': 'America/Argentina/Buenos_Aires',
    'America/Argentina/San_Luis': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Tucuman': 'America/Argentina/Buenos_Aires',
    'America/Argentina/Ushuaia': 'America/Argentina/Buenos_Aires',
    'America/Nipigon': 'America/Toronto',
    'Arctic/Longyearbyen': 'Europe/Oslo',
    'Asia/Famagusta': 'Asia/Nicosia',
    'Asia/Kuching': 'Asia/Kuala_Lumpur',
    'Europe/Busingen': 'Europe/Berlin',
    'Europe/Mariehamn': 'Europe/Helsinki'
}

#
# parameters for the spatial index
#
featureAreaRatioThreshold = 0.02 # 1% at zDepth 11 is ~150m
zDepth = 11 # minimum tile size is 1/(2^zdepth), amount of bits needed to store z index is 2*zDepth

# z-order curve coverage parameters
xStart = -180
xRange = 360
# cut out artic regions (starting at 60°S and 80°N), that saves about 23% z-order curve coverage which we
# can better use to increase precision in more relevant areas
yStart = -60
yRange = 140


# constants
LOG_CATEGORY = 'KI18n Data Generator'



# SPDX-FileCopyrightText: 2021 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: LGPL-2.0-or-later

from config import *
import io
import os
from qgis import *
from qgis.core import *

def tzIdToEnum(tzId):
    return tzId.replace('/', '_').replace('-', '_')

def normalizedTz(tzId):
    if tzId in TZID_MAP:
        return TZID_MAP[tzId]
    return tzId

def normalizedCountry(code):
    if code in ISO3166_1_MAP:
        return ISO3166_1_MAP[code]
    return code

# Generate IANA timezone names string table
# This allows us to reference timezones by a uint16_t in other data tables
class TimezoneStringTableTask(QgsTask):
    def __init__(self, context):
        super().__init__('Generate timezone string table', QgsTask.CanCancel)
        self.context = context

    def run(self):
        QgsMessageLog.logMessage('Generating timezone string table', LOG_CATEGORY, Qgis.Info)
        tzLayer = self.context['tzLayer']
        tzIds = set()
        for tz in tzLayer.getFeatures():
            tzIds.add(tz['tzid'])
        tzIds = list(tzIds)
        tzIds.sort()
        offsets = [0]

        out = open('../../data/timezone_name_table.cpp', 'w')
        out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated using QGIS - do not edit!
 */

static constexpr const char timezone_name_table[] =
""")
        for tzId in tzIds:
            out.write(f"    \"{tzId}\\0\"\n")
            offsets.append(offsets[-1] + len(tzId) + 1)
        out.seek(out.tell() - 1, io.SEEK_SET) # to make clang-format happy
        out.write(";\n")
        out.close()

        out = open('../../data/timezone_names_p.h', 'w')
        out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated using QGIS - do not edit!
 */

#ifndef TIMEZONE_NAMES_P_H
#define TIMEZONE_NAMES_P_H

#include <cstdint>

enum Tz : uint16_t {
""")
        for i in range(len(tzIds)):
            out.write(f"    {tzIdToEnum(tzIds[i])} = {offsets[i]},\n")
        out.write(f"    Undefined = {offsets[-1] - 1},\n") # points to the last null byte
        out.write("};\n\n#endif\n")
        out.close()
        return True

# Computes country/country subdivision to timezone mapping
class RegionToTimezoneMapTask(QgsTask):
    def __init__(self, context):
        super().__init__('Computing region to timezone mapping', QgsTask.CanCancel)
        self.context = context

    def run(self):
        QgsMessageLog.logMessage('Computing region to timezone mapping', LOG_CATEGORY, Qgis.Info)
        countryLayer = self.context['countryLayer']
        tzLayer = self.context['tzLayer']
        countryToTz = {}
        for country in countryLayer.getFeatures():
            countryCode = country['ISO3166-1']
            if not countryCode in countryToTz:
                countryToTz[countryCode] = set()
            countryGeom = country.geometry()
            countryArea = countryGeom.area()
            for tz in tzLayer.getFeatures():
                tzId = normalizedTz(tz['tzId'])
                if tz.geometry().intersects(countryGeom):
                    # filter out intersection noise along the boundaries
                    area = tz.geometry().intersection(countryGeom).area()
                    tzAreaRatio = area / tz.geometry().area()
                    countryAreaRatio = area / countryArea
                    if tzAreaRatio > 0.01 or countryAreaRatio > 0.1:
                        countryToTz[countryCode].add(tzId)

        out = open('../../data/country_timezone_map.cpp', 'w')
        out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated using QGIS - do not edit!
 */

#include "isocodes_p.h"
#include "mapentry_p.h"
#include "timezone_names_p.h"

static constexpr const MapEntry<uint16_t> country_timezone_map[] = {
""")
        countries = list(countryToTz)
        countries.sort()
        for country in countries:
            if len(countryToTz[country]) == 1:
                out.write(f"    {{IsoCodes::alpha2CodeToKey(\"{country}\"), Tz::{tzIdToEnum(list(countryToTz[country])[0])}}},\n")

        out.write("};\n")
        out.close()

        # for countries with more than one tz, match against subdivisions
        subdivToTz = {}
        subdivLayer = self.context['subdivLayer']
        tzLayer = self.context['tzLayer']
        for subdiv in subdivLayer.getFeatures():
            code = subdiv['ISO3166-2']
            country = code[:2]
            if len(countryToTz[country]) <= 1:
                continue
            if not code in subdivToTz:
                subdivToTz[code] = {}
            subdivGeom = subdiv.geometry()
            subdivArea = subdivGeom.area()
            for tz in tzLayer.getFeatures():
                tzId = normalizedTz(tz['tzId'])
                if tz.geometry().intersects(subdivGeom):
                    # filter out intersection noise along the boundaries
                    area = tz.geometry().intersection(subdivGeom).area()
                    tzAreaRatio = area / tz.geometry().area()
                    subdivAreaRatio = area / subdivArea
                    if tzAreaRatio > 0.01 or subdivAreaRatio > 0.1:
                        if not tzId in subdivToTz[code]:
                            subdivToTz[code][tzId] = area
                        else:
                            subdivToTz[code][tzId] += area

        out = open('../../data/subdivision_timezone_map.cpp', 'w')
        out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated using QGIS - do not edit!
 */

#include "isocodes_p.h"
#include "mapentry_p.h"
#include "timezone_names_p.h"

static constexpr const MapEntry<uint32_t> subdivision_timezone_map[] = {
""")
        subdivs = list(subdivToTz)
        subdivs.sort()
        for subdiv in subdivs:
            # sort by area, biggest one first
            tzs = list(subdivToTz[subdiv])
            tzs.sort(key = lambda x: subdivToTz[subdiv][x], reverse = True)
            for tz in tzs:
                out.write(f"    {{IsoCodes::subdivisionCodeToKey(\"{subdiv}\"), Tz::{tzIdToEnum(tz)}}},\n")
            if len(subdivToTz[subdiv]) == 0:
                out.write(f"    // {subdiv}\n")

        out.write("};\n")
        out.close()

        self.context['countryToTimezoneMap'] = countryToTz
        self.context['subdivisionToTimezoneMap'] = subdivToTz
        return True

# Compute timezone to country mapping
class TimezoneToCountryMapTask(QgsTask):
    def __init__(self, context):
        super().__init__('Computing timezone to country mapping', QgsTask.CanCancel)
        self.context = context

    def disambiguateCountries(self, countries):
        if len(countries) != 2:
            return countries
        for map in ISO3166_1_DISAMBIGUATION_MAP:
            if map[0] in countries and map[1] in countries:
                countries.remove(map[1])
                return countries
        return countries

    def run(self):
        QgsMessageLog.logMessage('Computing timezone to country mapping', LOG_CATEGORY, Qgis.Info)
        countryLayer = self.context['countryLayer']
        tzLayer = self.context['tzLayer']
        tzToCountry = {}
        for tz in tzLayer.getFeatures():
            tzId = tz['tzId'] # non-normalized!
            tzGeom = tz.geometry()
            tzArea = tzGeom.area()
            if not tz in tzToCountry:
                tzToCountry[tzId] = set()
            for country in countryLayer.getFeatures():
                code = normalizedCountry(country['ISO3166-1'])
                if code and country.geometry().intersects(tzGeom):
                    # filter out intersection noise along the boundaries
                    area = country.geometry().intersection(tzGeom).area()
                    tzAreaRatio = area / tzArea
                    countryAreaRatio = area / country.geometry().area()
                    if tzAreaRatio > 0.01 or countryAreaRatio > 0.1:
                        tzToCountry[tzId].add(code)

        out = open('../../data/timezone_country_map.cpp', 'w')
        out.write("""/*
 * SPDX-License-Identifier: ODbL-1.0
 * SPDX-FileCopyrightText: OpenStreetMap contributors
 *
 * Autogenerated using QGIS - do not edit!
 */

#include "isocodes_p.h"
#include "mapentry_p.h"
#include "timezone_names_p.h"

static constexpr const MapEntry<uint16_t> timezone_country_map[] = {
""")
        tzIds = list(tzToCountry)
        tzIds.sort()
        for tzId in tzIds:
            countries = self.disambiguateCountries(tzToCountry[tzId])
            if len(countries) == 1:
                out.write(f"    {{Tz::{tzIdToEnum(tzId)}, IsoCodes::alpha2CodeToKey(\"{list(countries)[0]}\")}},\n")
            else:
                out.write(f"    // Tz::{tzIdToEnum(tzId)}\n")
        out.write("};\n")
        out.close()
        return True



# SPDX-FileCopyrightText: 2021 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: LGPL-2.0-or-later

import os
import qgis
from LayerTasks import *
from SpatialIndexTasks import *
from TimezoneTableTasks import *
from config import *

class MainTask(QgsTask):
    def __init__(self, context):
        super().__init__('Generating geographic data for KI18N', QgsTask.CanCancel)
        self.context = context
        self.loadLayersTask = LoadLayersTask(context)
        self.addSubTask(self.loadLayersTask, [], QgsTask.ParentDependsOnSubTask)

        self.tzStringTableTask = TimezoneStringTableTask(context)
        self.addSubTask(self.tzStringTableTask, [self.loadLayersTask], QgsTask.ParentDependsOnSubTask)
        self.regionToTzMapTask = RegionToTimezoneMapTask(context)
        self.addSubTask(self.regionToTzMapTask, [self.loadLayersTask], QgsTask.ParentDependsOnSubTask)
        self.tzToCountryMapTask = TimezoneToCountryMapTask(context)
        self.addSubTask(self.tzToCountryMapTask, [self.loadLayersTask], QgsTask.ParentDependsOnSubTask)

        self.spatialIndexTask = SpatialIndexerTask(context, self.loadLayersTask)
        self.addSubTask(self.spatialIndexTask, [self.loadLayersTask, self.regionToTzMapTask], QgsTask.ParentDependsOnSubTask)

    def run(self):
        QgsMessageLog.logMessage('Generation completed.', LOG_CATEGORY, Qgis.Info)
        return True

# main
os.chdir(os.path.join(os.path.dirname(QgsProject.instance().fileName()), 'data'))
context = {}
task = MainTask(context)
QgsApplication.taskManager().addTask(task)



# SPDX-FileCopyrightText: 2021 Volker Krause <vkrause@kde.org>
# SPDX-License-Identifier: LGPL-2.0-or-later

import os
import requests
from qgis import *
from qgis.core import *
import time
import zipfile
from config import *

# Download and unpack Shapefiles
class LayerDownloadTask(QgsTask):
    def __init__(self, url, dest):
        super().__init__('Download Shapefile', QgsTask.CanCancel)
        self.url = url
        self.dest = dest

    def run(self):
        try:
            QgsMessageLog.logMessage(f"Downloading and unpacking {self.dest}...", LOG_CATEGORY, Qgis.Info)
            if not os.path.exists(self.dest):
                r = requests.get(self.url)
                if r.status_code < 400:
                    with open(self.dest, 'wb') as f:
                        f.write(r.content)
            with zipfile.ZipFile(self.dest, 'r') as z:
                z.extractall('.')
            QgsMessageLog.logMessage(f"Downloaded and unpacked {self.dest}.", LOG_CATEGORY, Qgis.Info)
        except Exception as e:
            QgsMessageLog.logMessage(f"Exception in task: {e}", LOG_CATEGORY, Qgis.Critical)
        return True


# Load and simplify Shapefile layers
# Simplification is done to massively speed up geometry intersection computation
# (for reference: the original KItinerary tz spatial index took 8h to compute without simplification,
# and about 15 minutes with a Douglas Peucker simplification with a 0.001 threshold, with no practical
# loss of precision
class LoadLayerTask(QgsTask):
    def __init__(self, url, fileName, context, layerName):
        super().__init__(f"Loading layer {fileName}", QgsTask.CanCancel)
        self.layer = None
        self.url = url
        self.fileName = fileName
        self.context = context
        self.layerName = layerName
        self.downloadTask = LayerDownloadTask(url, fileName)
        self.addSubTask(self.downloadTask, [], QgsTask.ParentDependsOnSubTask)

    def run(self):
        QgsMessageLog.logMessage(f"Simplifying layer {self.fileName}", LOG_CATEGORY, Qgis.Info)
        fullLayer = QgsVectorLayer(self.fileName, f"{self.fileName}-full-resolution", 'ogr')
        if not fullLayer.isValid():
            QgsMessageLog.logMessage(f"Failed to load layer {self.fileName}!", LOG_CATEGORY, Qgis.Critical)
        result = processing.run('qgis:simplifygeometries', {'INPUT': fullLayer, 'METHOD': 0, 'TOLERANCE': 0.001, 'OUTPUT': 'TEMPORARY_OUTPUT' })
        self.layer = result['OUTPUT']
        self.layer.setName(f"{self.fileName}-simplified")
        self.context[self.layerName] = self.layer
        QgsMessageLog.logMessage(f"Simplified layer {self.fileName}", LOG_CATEGORY, Qgis.Info)
        return True

    def finished(self, result):
        QgsProject.instance().addMapLayer(self.layer)


# Filter out too small elements in the ISO 3166-2 layer
class Iso3166_2FilterTask(QgsTask):
    def __init__(self, context):
        super().__init__('Filtering ISO 3166-2 layer', QgsTask.CanCancel)
        self.context = context

    def run(self):
        QgsMessageLog.logMessage('Filtering ISO 3166-2 layer', LOG_CATEGORY, Qgis.Info)
        subdivLayer = self.context['subdivLayer']
        toBeRemoved = []
        for feature in subdivLayer.getFeatures():
            # sic: the key is really "admin_leve" in the input file, due to length restrictions in the Shapefile...
            level = feature['admin_leve']
            country = feature['ISO3166-2'][:2]
            if not isinstance(level, str) or not isinstance(country, str):
                continue
            for filter in ISO3166_2_FILTER:
                if int(level) == filter['admin_level'] and country == filter['country']:
                    toBeRemoved.append(feature.id())
                    break
        subdivLayer.dataProvider().deleteFeatures(toBeRemoved)
        return True


# Setup all data layers we need
class LoadLayersTask(QgsTask):
    def __init__(self, context):
        super().__init__('Loading layers...', QgsTask.CanCancel)
        self.context = context
        self.tasks = [
            LoadLayerTask(TZDATA_URL, f"timezones.shapefile-{TZDATA_VERSION}.zip", context, 'tzLayer'),
            LoadLayerTask(ISO3166_1_URL, f"iso3166-1-boundaries.shp-{ISO3166_1_VERSION}.zip", context, 'countryLayer'),
            LoadLayerTask(ISO3166_2_URL, f"iso3166-2-boundaries.shp-{ISO3166_2_VERSION}.zip", context, 'subdivLayer')
        ]
        for task in self.tasks:
            self.addSubTask(task, [], QgsTask.ParentDependsOnSubTask)

        self.filterTask = Iso3166_2FilterTask(context)
        self.addSubTask(self.filterTask, [self.tasks[2]], QgsTask.ParentDependsOnSubTask)

    def run(self):
        return True
